%Calculate Linear IV regressions reported in Table V of Angrist Kreueger
%1991.  Here Consider same specifications as in Staiger and Stock (1997)
clear
rng(1);

coverage=.95;

%Set number of endogenous regressors
endog_reg=1;

beta_lower_bound=-.5;
beta_upper_bound=.5;

%Specify whether to use laptop or server specifications
laptop=0;
if laptop==1
    chain_draws=10^3;
    size_sim_draws=20;
    parfor_block_size=100;
    %parpool('local',2)
else
    chain_draws=10^6;
    %chain_draws=10^5;
    size_sim_draws=500;
    worker_count=24;
    parfor_block_size=100;
    while worker_count>=12
        try
            parpool('local',worker_count)
        end
        n = numlabs;
        if n==worker_count
            worker_count=0;
        else
            worker_count=worker_count-2;
        end
    end
end

count=1;
while count<endog_reg
    beta_lower_bound(end+1,1)=-2;
    beta_upper_bound(end+1,1)=2;
    count=count+1;
end

try
    load simrecord_QLR_1
end

if exist('n1')==0
    n1=0;
    
    %Use homoskedastic or heteroskedastic version of robust statistics
    heterosked_robust=1;
    
    %Choose which speficiation to use
    spec=2;
    
    load AK_3039_analysis_dataset;
    st_dataset(end,:)=[];
    
    AGEQ=double(st_dataset(1:end,2));
    AGEQSQ=double(st_dataset(1:end,28));
    EDUC=double(st_dataset(1:end,4));
    ENOCENT=double(st_dataset(1:end,5));
    ESOCENT=double(st_dataset(1:end,6));
    LWKLYWGE=double(st_dataset(1:end,8));
    MARRIED=double(st_dataset(1:end,9));
    MIDATL=double(st_dataset(1:end,10));
    MT=double(st_dataset(1:end,11));
    NEWENG=double(st_dataset(1:end,12));
    RACE=double(st_dataset(1:end,18));
    SMSA=double(st_dataset(1:end,19));
    SOATL=double(st_dataset(1:end,20));
    WNOCENT=double(st_dataset(1:end,23));
    WSOCENT=double(st_dataset(1:end,24));
    YR20_28=double(st_dataset(1:end,29:37));
    QTR120_129=double(st_dataset(1:end,43:52));
    QTR220_229=double(st_dataset(1:end,53:62));
    QTR320_329=double(st_dataset(1:end,63:72));
    QTR1_3=double(st_dataset(1:end,39:41));
    v17=double(st_dataset(1:end,16));
    
    SOB_index=[1 2 4 5 6 8 9 10 11 12 13 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 ...
        32 33 34 35 36 37 38 39 40 41 42 44 45 46 47 48 49 50 51 53 54 55]; %drop 56 to avoid multicollinartity
    SOB_dummies=zeros(size(QTR1_3,1),length(SOB_index));
    for n=1:length(SOB_index)
        SOB_dummies(:,n)=double(v17==SOB_index(n));
    end
    
    HS=(EDUC>=12);
    JR_College=(EDUC>=14);
    College=(EDUC>=16);
    Masters=(EDUC>=18);
    %%
    
    %Outcome variable is wages, endogenous variable is education, Z is
    %instruments, and W are controls
    Y=LWKLYWGE;
    if endog_reg==1
        X=EDUC;
    elseif endog_reg==2
        X=[EDUC HS];
    elseif endog_reg==3
        X=[EDUC HS JR_College];
    elseif endog_reg==4
        X=[EDUC HS JR_College College];
    elseif endog_reg==5
        X=[EDUC HS JR_College College Masters];
    end
    
    
    %Collect exogenous control variables and instruments.  Have checked that
    %results in 1-3 match corresponding results in Staiger and Stock.
    if spec==1
        Z=QTR1_3;
        W=[ones(length(Y),1) RACE MARRIED SMSA NEWENG MIDATL ENOCENT WNOCENT SOATL ESOCENT WSOCENT MT YR20_28];
    elseif spec==2
        W=[ones(length(Y),1) RACE MARRIED SMSA NEWENG MIDATL ENOCENT WNOCENT SOATL ESOCENT WSOCENT MT YR20_28];
        Z=[QTR120_129 QTR220_229 QTR320_329];
    elseif spec==3
        W=[ones(length(Y),1) RACE MARRIED SMSA NEWENG MIDATL ENOCENT WNOCENT SOATL ESOCENT WSOCENT MT AGEQ AGEQSQ YR20_28];
        Z=[QTR120_129 QTR220_229 QTR320_329];
        Z(:,end)=[];
        Z(:,end-1)=[];
    else
        W=[ones(length(Y),1) RACE MARRIED SMSA NEWENG MIDATL ENOCENT WNOCENT SOATL ESOCENT WSOCENT MT AGEQ AGEQSQ YR20_28 SOB_dummies];
        Z=[QTR120_129 QTR220_229 QTR320_329];
        Z(:,end)=[];
        Z(:,end-1)=[];
        Z=[Z repmat(SOB_dummies,1,3).*repmat(QTR1_3,1,size(SOB_dummies,2))];
    end
    
    %Drop rows with missing observations
    data=[Y X Z W];
    nanrows=isnan(sum(data,2));
    Y=Y(nanrows==0,:);
    X=X(nanrows==0,:);
    Z=Z(nanrows==0,:);
    W=W(nanrows==0,:);
    
    clear data SOB_index AGEQ AGEQSQ EDUC ENOCENT ESOCENT LWKLYWGE MARRIED QTR220_229 QTR1_3 st_dataset
    clear MIDATL MT NEWENG RACE SMSA SOATL WNOCENT WSOCENT YR20_28 QTR120_129 QTR320_329 SOB_dummies v17
    clear College JR_College JR_HS Junior Sophmore Masters nanrows
    
    %Run as 2SLS with included exogenous variables
    X_tilde=[X W];
    Z_tilde=[Z W];
    
    ZZ=Z_tilde'*Z_tilde;
    ZZ_inv=ZZ^-1;
    ZX=Z_tilde'*X_tilde;
    XZ=X_tilde'*Z_tilde;
    ZY=Z_tilde'*Y;
    
    beta_tilde=(XZ*ZZ_inv*ZX)^-1*XZ*ZZ_inv*ZY; %Matches estimate from AK code and paper
    
    WW=W'*W;
    WZ=W'*Z;
    WX=W'*X;
    WY=W'*Y;
    
    Y=Y-W*WW^-1*WY;
    X=X-W*WW^-1*WX;
    Z=Z-W*WW^-1*WZ;
    
    ZZ=Z'*Z;
    ZZ_inv=ZZ^-1;
    ZX=Z'*X;
    XZ=X'*Z;
    ZY=Z'*Y;
    
    instruments=size(Z,2);
    obs=length(Y);
    Omega=NaN((size(X,2)+1)*instruments);
    hmat=repmat(Y,1,instruments).*Z;
    Omega(1:instruments,1:instruments)=cov(hmat);
    clear W hmat X_tilde Z_tilde
    for m1=0:size(X,2)
        if m1==0
            hmat1=repmat(Y,1,instruments).*Z;
        else
            hmat1=repmat(X(:,m1),1,instruments).*Z;
        end
        for m2=1:size(X,2)
            hmat2=repmat(X(:,m2),1,instruments).*Z;
            Omega(m1*instruments+1:(m1+1)*instruments,m2*instruments+1:(m2+1)*instruments)...
                =1/obs*hmat1'*hmat2-mean(hmat1)*mean(hmat2)';
            Omega(m2*instruments+1:(m2+1)*instruments,m1*instruments+1:(m1+1)*instruments)...
                =1/obs*(hmat1'*hmat2-mean(hmat1)*mean(hmat2)')';
        end
    end
    
    clear hmat1 hmat2 X Y Z;
    Sigma_gg=@(b) kron([1;-b],eye(instruments))'*Omega*kron([1;-b],eye(instruments));
    Sigma_tg=@(b) Omega(instruments+1:end,1:instruments)...
        -Omega(instruments+1:end,instruments+1:end)*kron(b,eye(instruments));
    ZX_vec=reshape(ZX,length(ZX(1:end)),1);
    g=@(b) ZY-ZX*b;
    D=@(b) reshape(ZX_vec-Sigma_tg(b)*Sigma_gg(b)^-1*g(b),instruments,size(ZX,2));
    S=@(b) 1/obs*g(b)'*Sigma_gg(b)^-1*g(b);
    
    K_draws=sum(randn(size_sim_draws,size(ZX,2)).^2,2);
    J_draws=sum(randn(size_sim_draws,instruments-size(ZX,2)).^2,2);
    rk_grid=[0:.1:10 11:1:100 110:10:1000].^2;
    
    QCLR_crit=zeros(length(rk_grid),1);
    for m=1:length(rk_grid)
        r=rk_grid(m);
        QCLR_crit(m,1)=quantile(.5*(K_draws+J_draws-r+...
            ((K_draws+J_draws+r).^2-4*J_draws*r).^.5),coverage);
    end
    
    chain_store=zeros(chain_draws,size(ZX,2));
    S_store=ones(chain_draws,1)*1000;
    K_store=ones(chain_draws,1)*1000;
    rk_store=zeros(chain_draws,1);
    
    parfor m=2:chain_draws
        %for m=2:chain_draws
        if round(m*100/chain_draws)==m*100/chain_draws
            m/chain_draws
        end
        
        proposal=(rand(size(ZX,2),1)-.5).*(beta_upper_bound-beta_lower_bound)+(beta_upper_bound+beta_lower_bound)/2;
        
        A=kron([1;-proposal],eye(instruments));
        Sigma_temp=A'*Omega*A;
        Sigma_temp_inv=Sigma_temp^-1;
        Sigma_tg_temp=Omega(instruments+1:end,1:instruments)...
            -Omega(instruments+1:end,instruments+1:end)*kron(proposal,eye(instruments));
        g_temp=g(proposal);
        S_val=1/obs*g_temp'*Sigma_temp_inv*g_temp;
        S_store(m,1)=S_val;
        D_temp=reshape(ZX_vec-Sigma_tg_temp*Sigma_temp_inv*g_temp,instruments,size(ZX,2));
        
        K_store(m,1)=1/obs*g_temp'*Sigma_temp_inv*D_temp*(D_temp'*Sigma_temp_inv*D_temp)...
            ^-1*D_temp'*Sigma_temp_inv*g_temp;
        chain_store(m,:)=proposal;
        J_temp=S_store(m,1)-K_store(m,1);
        
        if size(ZX,2)==1;
            D_temp=D_temp';
        end
        Sigma_D_temp=Omega(instruments+1:end,instruments+1:end)-Sigma_tg_temp...
            *Sigma_temp_inv*Sigma_tg_temp';
        Sigma_D_temp_inv_rt=Sigma_D_temp^-.5;
        D_normalized=reshape(Sigma_D_temp_inv_rt*D_temp(1:end)',instruments,size(ZX,2));
        rk_store(m,1)=1/obs*min(eig(D_normalized'*D_normalized));
    end
    
    chain_truncate_S=chain_store(S_store<chi2inv(coverage,instruments),:);
    chain_truncate_K=chain_store(K_store<chi2inv(coverage,size(ZX,2)),:);
    chain_truncate_JK=chain_store((K_store<chi2inv(1-(1-coverage)*.8,size(ZX,2)))&((S_store-K_store)<chi2inv(1-(1-coverage)*.2,instruments-size(ZX,2))),:);
    
    S_min_ind=S_store==min(S_store);
    b_min_init=chain_store(S_min_ind==1,:)';
    [b_min, S_min]=fminsearch(S,b_min_init);
    %Calculate QLR statistic
    QLR_store=S_store-S_min;
    chain_truncate_QLR_init=chain_store(QLR_store<chi2inv(coverage,instruments),:);
    
    %Generate random draws used in critical value calculation
    xi_mat=randn(size_sim_draws,instruments);
    
    %Generate vector of starting values to use in optimization
    %vecmat=[beta_lower_bound:(beta_upper_bound-beta_lower_bound)/3:beta_upper_bound];
    grid_points=5;
    vecmat=[beta_lower_bound:(beta_upper_bound-beta_lower_bound)/(grid_points-1):beta_upper_bound];
    for n=2:length(beta_lower_bound)
        vecmat(n,:)=beta_lower_bound(n):((beta_upper_bound(n)-beta_lower_bound(n))/(grid_points-1)):beta_upper_bound(n);
    end
    
    if endog_reg==1
        M1=vecmat;
        startmat_base=M1(1:end)';
    elseif endog_reg==2
        [M1, M2]=ndgrid(vecmat(1,:),vecmat(2,:));
        startmat_base=[M1(1:end)' M2(1:end)'];
    elseif endog_reg==3
        [M1, M2, M3]=ndgrid(vecmat(1,:),vecmat(2,:),vecmat(3,:));
        startmat_base=[M1(1:end)' M2(1:end)' M3(1:end)'];
    elseif endog_reg==4
        [M1, M2, M3, M4]=ndgrid(vecmat(1,:),vecmat(2,:),vecmat(3,:),vecmat(4,:));
        startmat_base=[M1(1:end)' M2(1:end)' M3(1:end)' M4(1:end)'];
    elseif endog_reg==5
        [M1, M2, M3, M4, M5]=ndgrid(vecmat(1,:),vecmat(2,:),vecmat(3,:),vecmat(4,:),vecmat(5,:));
        startmat_base=[M1(1:end)' M2(1:end)' M3(1:end)' M4(1:end)' M5(1:end)'];
    end
    
    QLR_store=QLR_store(QLR_store<chi2inv(coverage,instruments),:);
    QLR_crit_store=zeros(size(chain_truncate_QLR_init,1),1);
    
    big_sigma_inv_mat_init=zeros(size(startmat_base,1)*instruments,instruments);
    for k=1:size(startmat_base,1)
        beta_temp=startmat_base(k,:)';
        A=kron([1;-beta_temp],eye(instruments));
        Sigma_temp=A'*Omega*A;
        Sigma_temp_inv=Sigma_temp^-1;
        big_sigma_inv_mat_init((k-1)*instruments+1:k*instruments,:)=Sigma_temp_inv;
    end
else
    load AK_QLR_1_1M_draws
end

tic;

while n1<=size(chain_truncate_QLR_init,1)/parfor_block_size
    parfor m=n1*parfor_block_size+1:...
            min((n1+1)*parfor_block_size+1,size(chain_truncate_QLR_init,1))
        %     for m=n1*parfor_block_size+1:...
        %             min((n1+1)*parfor_block_size+1,size(chain_truncate_QLR_init,1))
        m/(size(chain_truncate_QLR_init,1))
        beta0=chain_truncate_QLR_init(m,:)';
        %startmat=beta0';
        startmat=[beta0'; startmat_base];
        
        A=kron([1;-beta0],eye(instruments));
        Sigma_temp=A'*Omega*A;
        Sigma_temp_inv=Sigma_temp^-1;
        Sigma_temp_inv_rt=real(Sigma_temp_inv^.5);
        Sigma_temp_rt=Sigma_temp^.5;
        Sigma_tg_temp=Omega(instruments+1:end,1:instruments)...
            -Omega(instruments+1:end,instruments+1:end)*kron(beta0,eye(instruments));
        Big_sigma_inv_mat=[Sigma_temp_inv;big_sigma_inv_mat_init];
        
        g_temp=g(beta0);
        D_temp=reshape(ZX_vec-Sigma_tg_temp*Sigma_temp_inv*g_temp,instruments,size(ZX,2));
        D_temp_scaled=obs^-.5*D_temp;
        
        QLR_crit_store(m,1)=QLR_crit_calc_startsearch2(startmat,beta0,coverage,xi_mat,Sigma_temp_rt,D_temp_scaled,Sigma_temp_inv,Sigma_tg_temp,Omega,Big_sigma_inv_mat);
    end
    save AK_QLR_1_1M_draws
    n1=n1+1;
    save simrecord_QLR_1 n1
end
chain_truncate_QLR=chain_truncate_QLR_init(QLR_store<QLR_crit_store,:);


% poolobj = gcp('nocreate'); % If no pool, do not create new one.
% if isempty(poolobj)
%     poolsize = 0;
% else
%     poolsize = poolobj.NumWorkers
% end
% total_PI_runtime=toc*poolsize;
% average_PI_runtimes=toc/size(chain_truncate_all_LC,1)*poolsize;


clear bmat D_noise2 K_draws J_draws TJ TK
if laptop==1
    %delete(gcp)
else
    if endog_reg==1
        save AK_QLR_1_1M_draws
    elseif endog_reg==2
        save AK_QLR_2_1M_draws
    elseif endog_reg==3
        save AK_QLR_3_1M_draws
    elseif endog_reg==4
        save AK_QLR_4_1M_draws
    elseif endog_reg==5
        save AK_QLR_5_1M_draws
    end
    
    exit
end