% This code is due to David Wingate, and can be found at 
% http://www.mit.edu/~wingated/resources.html

% ---------------------------------------------------------------------------- 
% The rearrange operator rearranges a matrix in a form that is suitable
% for a rank-one approximation through kronecker products.
%

% ----------------------------------------------------------------------------

function [newA] = rearrange( A, m1, m2, n1, n2 )

% there should be m1*m2 rows and n1*n2 columns
[ h,w ] = size( A );
if ( h ~= m1*m2 )
   error( 'Incorrect blocking parameters m1 m2!' );
end;
if ( w ~= n1*n2 )
   error( 'Incorrect blocking parameters n1 n2!' );
end;

for j = 1:n1
  Aj = [];
  for i = 1:m1
    tmp = mblock( A, i, j, m2, n2 );
    tmpv = mtv( tmp )';
    Aj( i, 1:length(tmpv) ) = tmpv;
  end;

  [ th, tw ] = size( Aj );
  newA( (j-1)*th+1:j*th , 1:tw ) = Aj;
end;
