rng(1);

worker_count=24;
while worker_count>=12
    try
        parpool('local',worker_count)
    end
    n = numlabs;
    if n==worker_count
        worker_count=0;
    else
        worker_count=worker_count-2;
    end
end


%load Estim_params_rf_AUL

instruments=length(mu);

%Set number of simulation replications
%sims=500;
%sims=10;
%sims=2500;
%sims=10^4;
sims=5000;

%Set number of draws to use in calculating critical values
QLR_draws=1000;

%Set null parameter value
beta0=0;

%Set size of tests
alpha=0.05;
chi2crit=chi2inv(1-alpha,4);

Sigma_gg=Sigma(1:length(mu),1:length(mu));
mu=Sigma_gg^-.5*mu;
Sigma_standardized=kron(eye(2),Sigma_gg^-.5)*Sigma*kron(eye(2),Sigma_gg^-.5);
Sigma_gg=eye(instruments);
Sigma_tg=Sigma_standardized(length(mu)+1:end,1:length(mu));
Sigma_gt=Sigma_tg';
Sigma_tt=Sigma_standardized(length(mu)+1:end,length(mu)+1:end);
Sigma_D=Sigma_tt-Sigma_tg*Sigma_tg';
Sigmagg_inv=Sigma_gg^-1;
Sigma_rt=Sigma_standardized^.5;

%Set grid of values on which to calculate power
betagrid=(-6:.3:6)/norm(mu);

%set grid of values on which to evaluate test statistic
beta_base_grid=[0:.0025:3 3.0125:.0125:10 10.125:.125:20 20.5:.5:100 105:5:1000 1100:100:2000 2500:500:10^4];
beta_calc_grid=sort(unique([beta_base_grid -beta_base_grid betagrid]));

xi_mat=randn(instruments,QLR_draws);
xi_mat=Sigma_gg^.5*xi_mat;

g_noise=randn(sims,instruments);
D_noise=randn(sims,instruments);
g_noise=g_noise*Sigma_gg^.5;
D_noise=D_noise*Sigma_D^.5;

%Calculate covariances used to contruct moments
A_store=zeros(instruments*length(beta_calc_grid),instruments);
Sigma_theta_inv_rt_store=zeros(instruments*length(beta_calc_grid),instruments);
for m=1:length(beta_calc_grid)
    beta=beta_calc_grid(m);
    Sigma_thetatheta0=Sigma_gg-beta*Sigma_tg-beta0*Sigma_gt+beta*beta0*Sigma_tt;
    A_store((m-1)*instruments+1:m*instruments,:)=Sigma_thetatheta0;
    Sigma_thetatheta=Sigma_gg-beta*(Sigma_tg+Sigma_tg')+beta^2*Sigma_tt;
    Sigma_theta_inv_rt_store((m-1)*instruments+1:m*instruments,:)=real(Sigma_thetatheta^-.5);
end

QLR_stat=zeros(sims,length(betagrid));
QLR_crit=zeros(sims,length(betagrid));
pQLR_stat=zeros(sims,length(betagrid));
pQLR_crit=zeros(sims,length(betagrid));
AR_stat=zeros(sims,length(betagrid));
AR_stat_v2=zeros(sims,length(betagrid));
K_stat=zeros(sims,length(betagrid));
rk_stat=zeros(sims,length(betagrid));
for k=1:length(betagrid)
    beta=betagrid(k);
    %for n=1:sims;
    parfor n=1:sims
        [k/length(betagrid) n/sims]
        
        D=D_noise(n,:)';
        D=D+mu-Sigma_tg*mu*betagrid(k);
        g=g_noise(n,:)';
        g=g+mu*betagrid(k);
        G=D+Sigma_tg*g;
        
        K_stat(n,k)=g'*D*(D'*D)^-1*D'*g;
        AR_stat(n,k)=g'*g;
        rk_stat(n,k)=D'*Sigma_D^-1*D;
        
        moment_vec=repmat(g,length(beta_calc_grid),1)-kron(beta_calc_grid',G);
        h_vec=moment_vec-A_store*g;
        obj_vec=sum(Sigma_theta_inv_rt_store.*repmat(moment_vec,1,instruments),2);
        AR_vec=sum(reshape(obj_vec.^2,instruments,length(beta_calc_grid)));
        %AR_vec=(Summation_mat*obj_vec.^2)';
        
        AR_stat_v2(n,k)=AR_vec(1,beta_calc_grid==beta0);
        QLR_stat(n,k)=AR_vec(1,beta_calc_grid==beta0)-min(AR_vec(1,:),[],2);
        pQLR_stat(n,k)=AR_vec(1,beta_calc_grid==beta0)-...
            AR_vec(1,beta_calc_grid==betagrid(k));
        
        [QLR_crit_temp, pQLR_crit_temp]=QLR_crit_sim(h_vec,beta,beta0,beta_calc_grid,alpha,Sigma_theta_inv_rt_store,A_store,xi_mat);
        QLR_crit(n,k)=QLR_crit_temp;
        pQLR_crit(n,k)=pQLR_crit_temp;
    end
end

K_draws=sum(randn(10^4,1).^2,2);
J_draws=sum(randn(10^4,instruments-1).^2,2);
rk_grid=[0:.1:10 11:1:100 110:10:1000].^2;
QCLR_crit_vec=zeros(length(rk_grid),1);
for m=1:length(rk_grid)
    r=rk_grid(m);
    QCLR_crit_vec(m,1)=quantile(.5*(K_draws+J_draws-r+...
        ((K_draws+J_draws+r).^2-4*J_draws*r).^.5),1-alpha);
end
QCLR_crit=interp1(rk_grid,QCLR_crit_vec,rk_stat);
J_stat=AR_stat-K_stat;
QCLR_stat=.5*(K_stat+J_stat-rk_stat+...
    ((K_stat+J_stat+rk_stat).^2-4*J_stat.*rk_stat).^.5);


AR_power=mean(AR_stat>chi2inv(.95,instruments));
QLR_power=mean(QLR_stat>QLR_crit);
pQLR_power=mean(pQLR_stat>pQLR_crit);
K_power=mean(K_stat>chi2inv(.95,1));
pQLR_power(betagrid==beta0)=alpha;
QCLR_power=mean(QCLR_stat>QCLR_crit);

%plot(betagrid.*norm(mu),[AR_power' K_power' QCLR_power' QLR_power' pQLR_power']);
delete(gcp)